import functools
import time
from contextlib import ExitStack
from typing import Callable, Generic, TypeVar

__all__ = ('Instrumentation', 'Counter', 'Concurrency', 'EntryExit',)

IType = TypeVar('IType')


class Instrumentation(Generic[IType]):
    deinstrumentation = {}
    method: Callable
    wrap: Callable

    def __init__(self, target, methodname: str):
        assert isinstance(methodname, str)
        assert callable(getattr(target, methodname))
        self.target = target
        self.methodname = methodname

    def instrument(self, method, *args, **kwargs):
        raise NotImplementedError

    def __enter__(self: IType) -> IType:
        assert not hasattr(self, 'method')
        self.method = getattr(self.target, self.methodname)
        assert callable(self.method)

        @functools.wraps(self.method)
        def wrap(*args, **kwargs):
            return self.instrument(self.method, *args, **kwargs)

        self.wrap = wrap

        setattr(self.target, self.methodname, self.wrap)

        return self

    def schedule_deinstrumentation(self):
        self.deinstrumentation[self.wrap] = self.method
        del self.wrap
        del self.method

    def deinstrument(self):
        while (method := getattr(self.target, self.methodname)) in self.deinstrumentation:
            setattr(self.target, self.methodname, self.deinstrumentation.pop(method))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.schedule_deinstrumentation()
        self.deinstrument()

    def enter(self: IType, stack: ExitStack) -> IType:
        return stack.enter_context(self)


class Counter(Instrumentation):
    def __init__(self, target, methodname: str):
        assert isinstance(methodname, str)
        super().__init__(target, methodname)
        self.counter = 0

    def instrument(self, method, *args, **kwargs):
        self.counter += 1
        return method(*args, **kwargs)


class Concurrency(Instrumentation):
    start = time.time()

    def __init__(self, target, methodname: str):
        assert isinstance(methodname, str)
        super().__init__(target, methodname)
        self.concurrency = 0
        self.log: list[tuple[float, int]] = []

    def time(self) -> float:
        return time.time() - self.start

    def point(self) -> tuple[float, int]:
        return self.time(), self.concurrency

    async def instrument(self, method, *args, **kwargs):
        self.log.append(self.point())
        self.concurrency += 1
        self.log.append(self.point())
        result = await method(*args, **kwargs)
        self.log.append(self.point())
        self.concurrency -= 1
        self.log.append(self.point())
        return result


class EntryExit(Instrumentation):
    def __init__(self, target, methodname: str, point: Callable[[], tuple[float, int]]):
        assert callable(point)
        assert isinstance(methodname, str)
        super().__init__(target, methodname)
        self.point = point
        self.entry_log: list[tuple[float, int]] = []
        self.exit_log: list[tuple[float, int]] = []

    async def instrument(self, method, *args, **kwargs):
        self.entry_log.append(self.point())
        result = await method(*args, **kwargs)
        self.exit_log.append(self.point())
        return result
