import asyncio
import json
import os
import random
import shutil
import time
from contextlib import ExitStack

from nacl.signing import SigningKey

from plot import plot
from rainbowadn.chain import *
from rainbowadn.collection.trees.binary import *
from rainbowadn.core import *
from rainbowadn.nullability import *
from rainbowadn.testing.instrumentation import *
from rainbowadn.testing.resolvers import *
from rainbowadn.v13 import *


def get_dr() -> ExtendableResolver:
    dr = DictResolver()
    dr = DelayedResolver(dr, lambda: random.uniform(0.200, 0.500))
    # dr = CachingResolver(dr)
    return dr


def target_str(target) -> str:
    match target:
        case type(__name__=name):
            return name
        case object(__class__=type(__name__=name)):
            return name


def jsonify(instrumentation: Instrumentation) -> dict:
    prefix = f'{target_str(instrumentation.target)}:{instrumentation.methodname}'
    match instrumentation:
        case Counter(counter=counter):
            return {f'{prefix}:counter': counter}
        case Concurrency(log=log):
            return {f'{prefix}:concurrency': log}
        case EntryExit(entry_log=entry_log, exit_log=exit_log):
            return {
                f'{prefix}:entry': entry_log,
                f'{prefix}:exit': exit_log,
            }
        case _:
            return {}


async def main():
    set_gather_linear()
    bank: BankChain = BankChain.empty(ReductionChainMetaFactory().loose())
    key_0 = SigningKey.generate()
    transaction_0 = Transaction.make(
        [],
        [CoinData.of(Subject(key_0.verify_key), 100_000)],
        []
    )
    coin_0, coin_1 = await transaction_0.coins(MINT_CONST, NotNull(HashPoint.of(Subject(key_0.verify_key))))
    bank = await bank.adds(
        [
            transaction_0,
            Transaction.make(
                [coin_1],
                [CoinData.of(Subject(SigningKey.generate().verify_key), 10_000)],
                [key_0]
            ),
        ]
    )
    for _ in range(16):
        bank = await bank.adds(
            [
                Transaction.make(
                    [],
                    [CoinData.of(Subject(SigningKey.generate().verify_key), 0)] * 16,
                    []
                )
                for _ in range(16)
            ]
        )
    print('built')
    assert_true(await bank.verify())
    bank = BankChain.from_reference(
        ReductionChainMetaFactory(), await get_dr().migrate_resolved(bank.reference)
    )
    print('saved')
    set_gather_asyncio()
    with ExitStack() as stack:
        sleep_cc = Concurrency(DelayedResolver, 'sleep').enter(stack)
        avl_ex = EntryExit(ActiveBinaryTree, 'add', sleep_cc.point).enter(stack)
        assert_true(await bank.verify())
    print(Instrumentation.deinstrumentation)
    fn = f'trace/{int(time.time())}-{os.urandom(2).hex()}.json'
    with open(fn, 'w') as file:
        json.dump(
            jsonify(sleep_cc) | jsonify(avl_ex),
            file
        )
    shutil.copy(fn, f'trace/latest.json')
    plot()


if __name__ == '__main__':
    asyncio.run(main())
